%% Funktion zum Lösen von Ausgleichsproblemen anhand gegebener Werte
% Dominik Zobel - 2011

% Benötigt: Eingabewerte zur Erzeugung des Datensatzes "x"
%           Zu untersuchender Datensatz "y"
%           Gewünschtes Ausgangsformat Polynomial (p)/Exponential (e) "form"
%           Gewünschte Anzahl/Art der Koeffizienten (einer bis drei) "coef"
% Optional: Schreibe Infotext in die Ausgabe, wenn folgende Variable übergeben wird/existiert "verbosity"
% Übergibt: Berechnete Koeffizienten in einem Vektor "least_squares_coef"
%           Status der Bearbeitung "statusflag"

function [statusflag,least_squares_coef] = least_squares (x,y,form,coef,verbosity)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% -- Initialisierungen und Überprüfungen -- %%

% Initialisieren der Übergabevariablen
statusflag=1;                    % Wird am Ende des Skripts (nach erfolgreichem Durchlaufen) auf 0 gesetzt
least_squares_coef=[];

% Überprüfen, ob genügend Werte übergeben wurden
if ((~exist('x','var')) || (~exist('y','var')) || (~exist('form','var')) || (~exist('coef','var')))
   disp(['Abbruch [least_squares.m]: Nicht genügend Übergabevariablen']);
   clear x y form coef verbosity
   return;
end

if ((min(size(x)) ~= 1) || (min(size(y)) ~= 1))
   disp(['Abbruch [least_squares.m]: x und y müssen Vektoren sein']);
   clear x y form coef verbosity
   return;
end

% Alles als Spaltenvektor schreiben, so dass später die Ausgleichsprobleme mit Vektorrechnung geschehen kann
if (size(x,1) ~= 1)
   x=x';
end

if (size(y,1) ~= 1)
   y=y';
end

% Für das Ausgangsformat und die Anzahl/Art der Koeffizienten wird folgende Betrachtung gewählt:
% Aktuell verfügbar sind: - Lösen eines polynomialen Ausgleichsproblems bis zu 3 Koeffizienten (p)
%                              Dabei hat das Ausgleichspolynom die Form   y = a*x^alpha + b*x^beta + c*x^gamma
%                              Die Zuordnung der Koeffizienten ist beliebig
%                              "form" (p) und "coef" ([0 3]) würde also die beste Annäherung einer
%                              Funktion   y = a*x^0 + b*x^3 ermitteln und einen Vektor [a b] zurückgeben
%                         - Lösen eines exponentiellen Ausgleichsproblems (e)
%                              Hier ist die Form   y = b*exp( -c*x)
%                              Die Zuordnung lautet: b (0) ; c (1)
%                              Mit den Variablen "form" (e) und "coef" ([0 1]) wird die Funktion
%                              y = b*exp( -c*x) verwendet und der Vektor [b c] zurückgegeben.
%                              Um einen Grenzwert gegen Unendlich ungleich Null zu berücksichtigen, also
%                              eine Gleichung der Form   y = a + b*exp( -c*x), muss der a-Anteil im VORFELD
%                              abgezogen werden und nach der Berechnung von b und c wieder in Betracht gezogen
%                              werden, da dieses Skript nur die Form y = b*exp( -c*x) unterstützt.

done=0;

if (form == 'p')
   % Überprüfung, ob "coef" einen gültigen Wert hat
   if ((max(size(coef)) > 3) || (min(size(coef)) == 0))
      disp(['Abbruch [least_squares.m]: Ein bis drei Koeffizienten angeben']);
      clear done x y form coef
      return;
   end

   % Je nach Anzahl an Einträgen in "coef" kann eine kurze/effizientere Berechnung der Koeffizienten durchgeführt werden
   switch max(size(coef))
      case 1
         alpha=coef;

         % Wie auch bei den anderen Fällen wird eine direkte Gleichung zum Lösen der Koeffizienten benutzt
         least_squares_coef=(x.^alpha*y')/(x.^alpha*(x.^alpha)');

         if (exist('verbosity','var'))
            disp(['y=',num2str(least_squares_coef),'*x^',num2str(alpha)]);
         end
         clear alpha

      case 2
         alpha=coef(1);
         beta=coef(2);
         a=(x.^beta*(x.^beta)'*x.^alpha*y'-x.^alpha*(x.^beta)'*x.^beta*y')/(x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'-(x.^alpha*(x.^beta)')^2);
         b=(x.^alpha*(x.^alpha)'*x.^beta*y'-x.^alpha*(x.^beta)'*x.^alpha*y')/(x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'-(x.^alpha*(x.^beta)')^2);

         least_squares_coef=[a b];
         if (exist('verbosity','var'))
            disp(['y=',num2str(least_squares_coef(1)),'*x^',num2str(alpha),'+',num2str(least_squares_coef(2)),'*x^',num2str(beta)]);
         end
         clear a alpha b beta

      case 3
         alpha=coef(1);
         beta=coef(2);
         gamma=coef(3);
         a=(x.^alpha*y'*x.^beta*(x.^beta)'*x.^gamma*(x.^gamma)'-x.^alpha*y'*(x.^beta*(x.^gamma)')^2+x.^beta*y'*x.^alpha*(x.^gamma)'*x.^beta*(x.^gamma)'-x.^beta*y'*x.^alpha*(x.^beta)'*x.^gamma*(x.^gamma)'+x.^gamma*y'*x.^alpha*(x.^beta)'*x.^beta*(x.^gamma)'-x.^gamma*y'*x.^alpha*(x.^gamma)'*x.^beta*(x.^beta)')/(x.^alpha*(x.^beta)'*x.^alpha*(x.^gamma)'*x.^beta*(x.^gamma)'*2-x.^alpha*(x.^alpha)'*(x.^beta*(x.^gamma)')^2-x.^beta*(x.^beta)'*(x.^alpha*(x.^gamma)')^2-x.^gamma*(x.^gamma)'*(x.^alpha*(x.^beta)')^2+x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'*x.^gamma*(x.^gamma)');
         b=(x.^alpha*y'*x.^alpha*(x.^gamma)'*x.^beta*(x.^gamma)'-x.^alpha*y'*x.^alpha*(x.^beta)'*x.^gamma*(x.^gamma)'+x.^beta*y'*x.^alpha*(x.^alpha)'*x.^gamma*(x.^gamma)'-x.^beta*y'*(x.^alpha*(x.^gamma)')^2+x.^gamma*y'*x.^alpha*(x.^beta)'*x.^alpha*(x.^gamma)'-x.^gamma*y'*x.^alpha*(x.^alpha)'*x.^beta*(x.^gamma)')/(x.^alpha*(x.^beta)'*x.^alpha*(x.^gamma)'*x.^beta*(x.^gamma)'*2-x.^alpha*(x.^alpha)'*(x.^beta*(x.^gamma)')^2-x.^beta*(x.^beta)'*(x.^alpha*(x.^gamma)')^2-x.^gamma*(x.^gamma)'*(x.^alpha*(x.^beta)')^2+x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'*x.^gamma*(x.^gamma)');
         c=(x.^alpha*y'*x.^alpha*(x.^beta)'*x.^beta*(x.^gamma)'-x.^alpha*y'*x.^beta*(x.^beta)'*x.^alpha*(x.^gamma)'+x.^beta*y'*x.^alpha*(x.^beta)'*x.^alpha*(x.^gamma)'-x.^beta*y'*x.^alpha*(x.^alpha)'*x.^beta*(x.^gamma)'+x.^gamma*y'*x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'-x.^gamma*y'*(x.^alpha*(x.^beta)')^2)/(x.^alpha*(x.^beta)'*x.^alpha*(x.^gamma)'*x.^beta*(x.^gamma)'*2-x.^alpha*(x.^alpha)'*(x.^beta*(x.^gamma)')^2-x.^beta*(x.^beta)'*(x.^alpha*(x.^gamma)')^2-x.^gamma*(x.^gamma)'*(x.^alpha*(x.^beta)')^2+x.^alpha*(x.^alpha)'*x.^beta*(x.^beta)'*x.^gamma*(x.^gamma)');

         least_squares_coef=[a b c];
         if (exist('verbosity','var'))
            disp(['y=',num2str(least_squares_coef(1)),'*x^',num2str(alpha),'+',num2str(least_squares_coef(2)),'*x^',num2str(beta),'+',num2str(least_squares_coef(3)),'*x^',num2str(gamma)]);
         end
         clear a alpha b beta c gamma

      otherwise
         % Sollte aufgrund der Überprüfung von "coef" niemals aufgerufen werden
         disp(['Abbruch [least_squares.m]: coef=',num2str(coef),' nicht implementiert']);
         return;

   end
   % Hier angekommen sollte alles gut sein
   done=1;
end

if (form == 'e')
   % Es werden also die Koeffizienten einer Exponentialfunktion gesucht. Hier gibt es drei Möglichkeiten:
   % - coef=[ 0 ] -> Suche nach b, c=1 : y = b*e^( -  x)
   % - coef=[ 1 ] -> Suche nach c, b=1 : y =   e^( -c*x)
   % - coef=[0 1] -> Suche nach b und c: y = b*e^( -c*x)

   % Überprüfung, ob die richtigen "coef" Werte übergeben wurden
   if ((coef ~= 0) && (coef ~= 1) && (coef ~= [0 1]))
      disp(['Abbruch [least_squares.m]: Ein oder zwei Koeffizienten [0 1] für "coef" angeben']);
      clear done x y form coef
      return;
   end

   % Überprüfen, ob ein rein reelles Ergebnis existieren kann
   if (min(y) <= 0)
      disp(['Abbruch [least_squares.m]: Keine reale Lösung möglich']);
      return;
   end

   berechnet=0;

   % Für den Fall, dass der b-Wert für c=1 gesucht ist, kann die folgende Formel verwendet werden
   if (coef == 0)
      b=exp(1/size(y,2)*(sum(log(y)+x)));
      c=1;
      least_squares_coef=b;
      berechnet=1;
   end

   % Oder falls c gesucht wird für b=1
   if (coef == 1)
      b=1;
      c=-sum(log(y))/sum(x);
      least_squares_coef=c;
      berechnet=1;
   end

   % Beim Standardfall werden beide Parameter gesucht - hier in zwei unabhängigen Formeln
   if (coef == [0 1])
      b=exp((sum(x)*(x*log(y)')-x*x'*sum(log(y)))/(sum(x)*sum(x)-size(y,2)*x*x'));
      c=-(x*log(y)'-1/size(y,2)*sum(x)*sum(log(y)))/(x*x'-1/size(y,2)*sum(x)*sum(x));
      least_squares_coef=[b c];
      berechnet=1;
   end

   if (berechnet == 0)
      disp(['Abbruch [least_squares.m]: Nichts berechnet']);
      return;
   end

   if (exist('verbosity','var'))
      disp(['y=',num2str(b),'*e^(-(',num2str(c),')*x)']);
   end

   clear b c
   done=1;
end

if (done ~= 1)
   disp(['Abbruch [least_squares.m]: Ungültiger Wert für "form"']);
   clear done x y form coef
   return;
end

% Alles erledigt
statusflag=0;